@@ -2,18 +2,30 @@ use crate::comparator::FeatureComparator;
 use crate::nnnoiseless_fork::{self, DenoiseFeatures};
 use crate::wakeword::{Wakeword, WakewordModel};
 use hound::WavReader;
-use log::{debug, info, warn};
+use log::{debug, warn};
 use rubato::{FftFixedInOut, Resampler};
 use savefile::{load_file, load_from_mem, save_file, save_to_mem};
 use std::collections::HashMap;
 use std::fs::File;
 use std::io::BufReader;
 use std::path::Path;
+use std::sync::Arc;
 use std::thread::{self, JoinHandle};
 use std::time::SystemTime;
 static INTERNAL_SAMPLE_RATE: usize = 48000;
+/// Allowed vad modes
 pub type VadMode = webrtc_vad::VadMode;
+/// Allowed wav sample formats
 pub type SampleFormat = hound::SampleFormat;
+/// Use this struct to configure and build your wakeword detector.
+/// ```
+/// let mut word_detector = detector_builder
+///     .set_threshold(0.5)
+///     .set_sample_rate(16000)
+///     .set_eager_mode(true)
+///     .set_single_thread(true)
+///     .build();
+/// ```
 pub struct WakewordDetectorBuilder {
     sample_rate: Option<usize>,
     sample_format: Option<SampleFormat>,
@@ -23,6 +35,7 @@ pub struct WakewordDetectorBuilder {
     vad_delay: Option<u16>,
     eager_mode: bool,
     single_thread: bool,
+    averaged_threshold: Option<f32>,
     threshold: Option<f32>,
     comparator_band_size: Option<usize>,
     comparator_ref: Option<f32>,
@@ -41,6 +54,7 @@ impl WakewordDetectorBuilder {
             vad_delay: None,
             vad_sensitivity: None,
             threshold: None,
+            averaged_threshold: None,
             comparator_band_size: None,
             comparator_ref: None,
         }
@@ -56,48 +70,122 @@ impl WakewordDetectorBuilder {
             self.get_vad_delay(),
             self.get_vad_sensitivity(),
             self.get_threshold(),
+            self.get_averaged_threshold(),
             self.get_comparator_band_size(),
             self.get_comparator_ref(),
         )
     }
-    pub fn set_threshold(&mut self, value: f32) {
+    /// Configures the detector threshold,
+    /// is the min score (in range 0. to 1.) that some of
+    /// the wakeword templates should obtain to trigger a detection.
+    ///
+    /// Defaults to 0.5, wakeword defined value takes prevalence if present.
+    pub fn set_threshold(&mut self, value: f32) -> &mut Self {
         assert!(value >= 0. || value <= 1.);
         self.threshold = Some(value);
+        self
     }
-    pub fn set_bits_per_sample(&mut self, value: u16) {
+    /// Configures the detector threshold,
+    /// is the min score (in range 0. to 1.) that  
+    /// the averaged wakeword template should obtain to allow
+    /// to continue with the detection. This way it can prevent to
+    /// run the comparison of the current frame against each of the wakeword templates.
+    /// If set to 0. this functionality is disabled.
+    ///
+    /// Defaults to half of the configured threshold, wakeword defined value takes prevalence if present.
+    pub fn set_averaged_threshold(&mut self, value: f32) -> &mut Self {
+        assert!(value >= 0. || value <= 1.);
+        self.averaged_threshold = Some(value);
+        self
+    }
+    /// Configures the detector expected bit per sample for the audio chunks to process.
+    ///
+    /// Defaults to 16; Allowed values: 8, 16, 24, 32
+    pub fn set_bits_per_sample(&mut self, value: u16) -> &mut Self {
         self.bits_per_sample = Some(value);
+        self
     }
-    pub fn set_sample_rate(&mut self, value: usize) {
+    /// Configures the detector expected sample rate for the audio chunks to process.
+    ///
+    /// Defaults to 16000
+    pub fn set_sample_rate(&mut self, value: usize) -> &mut Self {
         self.sample_rate = Some(value);
+        self
     }
-    pub fn set_sample_format(&mut self, value: SampleFormat) {
+    /// Configures the detector expected sample format for the audio chunks to process.
+    ///
+    /// Defaults to int
+    pub fn set_sample_format(&mut self, value: SampleFormat) -> &mut Self {
         self.sample_format = Some(value);
+        self
     }
-    pub fn set_comparator_band_size(&mut self, value: usize) {
+    /// Configures the band-size for the comparator used to match the samples.
+    ///
+    /// Defaults to 6
+    pub fn set_comparator_band_size(&mut self, value: usize) -> &mut Self {
         self.comparator_band_size = Some(value);
+        self
     }
-    pub fn set_comparator_ref(&mut self, value: f32) {
+    /// Configures the reference for the comparator used to match the samples.
+    ///
+    /// Defaults to 0.22
+    pub fn set_comparator_ref(&mut self, value: f32) -> &mut Self {
         self.comparator_ref = Some(value);
+        self
     }
-    pub fn set_eager_mode(&mut self, value: bool) {
+    /// Enables eager mode.
+    /// Terminate the detection as son as one result is above the score,
+    /// instead of wait to see if the next frame has a higher score.
+    ///
+    /// Recommended for real usage.
+    ///
+    /// Defaults to false.
+    pub fn set_eager_mode(&mut self, value: bool) -> &mut Self {
         self.eager_mode = value;
+        self
     }
-    pub fn set_single_thread(&mut self, value: bool) {
+    /// Unless enabled the comparison against multiple wakewords run
+    /// in separate threads.
+    ///
+    /// Defaults to false.
+    ///
+    /// Only applies when more than a wakeword is loaded.
+    pub fn set_single_thread(&mut self, value: bool) -> &mut Self {
         self.single_thread = value;
+        self
     }
-    pub fn set_vad_delay(&mut self, value: u16) {
+    /// Seconds to disable the vad detector after voice is detected.
+    ///
+    /// Defaults to 3.
+    ///
+    /// Only applies if vad is enabled.
+    pub fn set_vad_delay(&mut self, value: u16) -> &mut Self {
         self.vad_delay = Some(value);
+        self
     }
-    pub fn set_vad_sensitivity(&mut self, value: f32) {
+    /// Voice/silence ratio in the last second to consider voice detected.
+    ///
+    /// Defaults to 0.5.
+    ///
+    /// Only applies if vad is enabled.
+    pub fn set_vad_sensitivity(&mut self, value: f32) -> &mut Self {
         assert!(value >= 0. || value <= 1.);
         self.vad_sensitivity = Some(value);
+        self
     }
-    pub fn set_vad_mode(&mut self, value: VadMode) {
+    /// Use a vad detector to reduce computation on absence of voice sound.
+    ///
+    /// Unless specified the vad detector is disabled.
+    pub fn set_vad_mode(&mut self, value: VadMode) -> &mut Self {
         self.vad_mode = Some(value);
+        self
     }
     fn get_threshold(&self) -> f32 {
         self.threshold.unwrap_or(0.5)
     }
+    fn get_averaged_threshold(&self) -> f32 {
+        self.averaged_threshold.unwrap_or(self.get_threshold() / 2.)
+    }
     fn get_sample_rate(&self) -> usize {
         self.sample_rate.unwrap_or(INTERNAL_SAMPLE_RATE)
     }
@@ -108,7 +196,7 @@ impl WakewordDetectorBuilder {
         self.bits_per_sample.unwrap_or(16)
     }
     fn get_comparator_band_size(&self) -> usize {
-        self.comparator_band_size.unwrap_or(6)
+        self.comparator_band_size.unwrap_or(11)
     }
     fn get_comparator_ref(&self) -> f32 {
         self.comparator_ref.unwrap_or(0.22)
@@ -137,17 +225,37 @@ impl WakewordDetectorBuilder {
         }
     }
 }
+/// This struct manages the wakeword generation and the spotting functionality.
+///
+/// ```
+/// // assuming the audio input format match the detector defaults
+/// let mut word_detector = detector_builder.build();
+/// // load and enable a wakeword
+/// word_detector.add_keyword_from_model("./model.rpw", true)?;
+/// let mut frame_buffer: Vec<i32> = vec![0; word_detector.get_samples_per_frame()];
+/// while true {
+///     // fill the buffer with new samples...
+///     let detection = word_detector.process(frame_buffer);
+///     if detection.is_some() {
+///         println!(
+///             "Detected '{}' with score {}!",
+///             detection.unwrap().wakeword,
+///             detection.unwrap().score,
+///         );
+///     }
+/// }
+/// ```
 pub struct WakewordDetector {
     // input options
     sample_format: SampleFormat,
     bits_per_sample: u16,
     // detection options
     threshold: f32,
-    comparator_band_size: usize,
-    comparator_ref: f32,
+    averaged_threshold: f32,
     eager_mode: bool,
     single_thread: bool,
     resampler: Option<FftFixedInOut<f32>>,
+    comparator: Arc<FeatureComparator>,
     vad_detector: Option<webrtc_vad::Vad>,
     vad_delay: u16,
     vad_sensitivity: f32,
@@ -157,15 +265,19 @@ pub struct WakewordDetector {
     min_frames: usize,
     max_frames: usize,
     frames: Vec<Vec<f32>>,
-    keywords: HashMap<String, Wakeword>,
+    wakewords: HashMap<String, Wakeword>,
     result_state: Option<DetectedWakeword>,
     extractor: DenoiseFeatures,
     resampler_out_buffer: Option<Vec<Vec<f32>>>,
+    vad_enabled: bool,
     voice_detections: Vec<bool>,
     voice_detection_time: SystemTime,
     audio_cache: Vec<Vec<f32>>,
 }
 impl WakewordDetector {
+    /// Creates a new WakewordDetector.
+    ///
+    /// It is recommended to use the WakewordDetectorBuilder struct instead.
     pub fn new(
         // input options
         sample_rate: usize,
@@ -178,6 +290,7 @@ impl WakewordDetector {
         vad_delay: u16,
         vad_sensitivity: f32,
         threshold: f32,
+        averaged_threshold: f32,
         comparator_band_size: usize,
         comparator_ref: f32,
     ) -> Self {
@@ -201,18 +314,18 @@ impl WakewordDetector {
         };
         let detector = WakewordDetector {
             threshold,
+            averaged_threshold,
             sample_format,
             bits_per_sample,
             samples_per_frame,
             frames: Vec::new(),
-            keywords: HashMap::new(),
+            wakewords: HashMap::new(),
             buffering: true,
             min_frames: 9999,
             max_frames: 0,
             result_state: None,
             extractor: DenoiseFeatures::new(),
-            comparator_band_size,
-            comparator_ref,
+            comparator: Arc::new(FeatureComparator::new(comparator_band_size, comparator_ref)),
             resampler_out_buffer: if resampler.is_some() {
                 Some(resampler.as_ref().unwrap().output_buffer_allocate())
             } else {
@@ -223,6 +336,7 @@ impl WakewordDetector {
             single_thread,
             voice_detections: Vec::with_capacity(100),
             audio_cache: Vec::with_capacity(100),
+            vad_enabled: false,
             vad_detector,
             vad_delay,
             vad_sensitivity,
@@ -230,72 +344,72 @@ impl WakewordDetector {
         };
         detector
     }
+    /// Loads a wakeword from its model bytes.
     pub fn add_keyword_from_model_bytes(
         &mut self,
         bytes: Vec<u8>,
-        enable_average: bool,
         enabled: bool,
     ) -> Result<(), String> {
         let model: WakewordModel = load_from_mem(&bytes, 0).or(Err("Unable to load model data"))?;
-        self.add_keyword_from_model(model, enable_average, enabled)
+        self.add_keyword_from_model(model, enabled)
     }
+    /// Loads a wakeword from its model path.
     pub fn add_keyword_from_model_file(
         &mut self,
         path: String,
-        enable_average: bool,
         enabled: bool,
     ) -> Result<(), String> {
         let model: WakewordModel = load_file(path, 0).or(Err("Unable to load model data"))?;
-        self.add_keyword_from_model(model, enable_average, enabled)
-    }
-    fn add_keyword_from_model(
-        &mut self,
-        model: WakewordModel,
-        enable_average: bool,
-        enabled: bool,
-    ) -> Result<(), String> {
-        let keyword = model.keyword.clone();
-        let wakeword = Wakeword::from_model(model, enable_average, enabled);
-        self.update_detection_frame_size(wakeword.get_min_frames(), wakeword.get_max_frames());
-        self.keywords.insert(keyword, wakeword);
-        Ok(())
-    }
-    pub fn generate_wakeword_model_file(&self, name: String, path: String) -> Result<(), String> {
-        let model = self.get_wakeword_model(&name)?;
-        save_file(path, 0, &model).or(Err(String::from("Unable to generate file")))
+        self.add_keyword_from_model(model, enabled)
     }
+    /// Generates the model file bytes from a loaded a wakeword.
     pub fn generate_wakeword_model_bytes(&self, name: String) -> Result<Vec<u8>, String> {
         let model = self.get_wakeword_model(&name)?;
         save_to_mem(0, &model).or(Err(String::from("Unable to generate model bytes")))
     }
-    fn update_detection_frame_size(&mut self, min_frames: usize, max_frames: usize) {
-        self.min_frames = std::cmp::min(self.min_frames, min_frames);
-        self.max_frames = std::cmp::max(self.max_frames, max_frames);
+    /// Generates a model file from a loaded a wakeword on the desired path.
+    pub fn generate_wakeword_model_file(&self, name: String, path: String) -> Result<(), String> {
+        let model = self.get_wakeword_model(&name)?;
+        save_file(path, 0, &model).or(Err(String::from("Unable to generate file")))
     }
+    /// Adds a wakeword using wav samples.
+    /// 
+    /// ```
+    /// let mut word_detector = detector_builder.build();
+    /// word_detector.add_keyword(
+    ///     model_name.clone(),
+    ///     enabled,
+    ///     averaged_threshold,
+    ///     threshold,
+    ///     sample_paths,
+    /// );
+    /// // Save as model file
+    /// word_detector.generate_wakeword_model_file(model_name.clone(), model_path)?;
+    /// ```
     pub fn add_keyword(
         &mut self,
-        keyword: String,
-        enable_average: bool,
+        name: String,
         enabled: bool,
+        averaged_threshold: Option<f32>,
         threshold: Option<f32>,
-        templates: Vec<String>,
+        sample_paths: Vec<String>,
     ) {
-        info!(
-            "Adding keyword \"{}\" (templates: {:?})",
-            keyword, templates
+        debug!(
+            "Adding keyword \"{}\" (sample paths: {:?})",
+            name, sample_paths
         );
-        if self.keywords.get_mut(&keyword).is_none() {
-            self.keywords.insert(
-                keyword.clone(),
-                Wakeword::new(enable_average, enabled, threshold),
+        if self.wakewords.get_mut(&name).is_none() {
+            self.wakewords.insert(
+                name.clone(),
+                Wakeword::new(enabled, averaged_threshold, threshold),
             );
         }
         let mut min_frames: usize = 0;
         let mut max_frames: usize = 0;
-        for template in templates {
+        for template in sample_paths {
             match self.extract_features_from_file(template) {
                 Ok(features) => {
-                    let word = self.keywords.get_mut(&keyword).unwrap();
+                    let word = self.wakewords.get_mut(&name).unwrap();
                     word.add_features(features.to_vec());
                     min_frames = if min_frames == 0 {
                         word.get_min_frames()
@@ -311,20 +425,31 @@ impl WakewordDetector {
         }
         self.update_detection_frame_size(min_frames, max_frames);
     }
-    fn get_wakeword_model(&self, name: &String) -> Result<WakewordModel, String> {
-        let keyword = self.keywords.get(name);
-        if keyword.is_none() {
-            Err(String::from("Missing wakeword"))
-        } else {
-            let features = keyword.unwrap().get_templates();
-            let model =
-                WakewordModel::new(name.clone(), features, keyword.unwrap().get_threshold());
-            Ok(model)
-        }
-    }
+    /// Process i32 audio chunks.
+    /// 
+    /// Asserts that the audio chunk length should match the return
+    /// of the get_samples_per_frame method.
+    /// 
+    /// Assumes sample rate match the configured for the detector.
+    /// 
+    /// Asserts that detector bits_per_sample is one of: 8, 16, 24, 32.
+    /// 
+    /// Asserts that detector sample_format is 'int'.
+    /// 
+    /// It's an alias for the process_i32 method.
     pub fn process(&mut self, audio_chunk: &[i32]) -> Option<DetectedWakeword> {
         self.process_i32(audio_chunk)
     }
+    /// Process i8 audio chunks.
+    /// 
+    /// Asserts that the audio chunk length should match the return
+    /// of the get_samples_per_frame method.
+    /// 
+    /// Assumes sample rate match the configured for the detector.
+    /// 
+    /// Asserts that detector bits_per_sample is 8.
+    /// 
+    /// Asserts that detector sample_format is 'int'.
     pub fn process_i8(&mut self, audio_chunk: &[i8]) -> Option<DetectedWakeword> {
         assert!(self.bits_per_sample == 8);
         self.process_int(
@@ -334,6 +459,16 @@ impl WakewordDetector {
                 .collect::<Vec<_>>(),
         )
     }
+    /// Process i16 audio chunks.
+    /// 
+    /// Asserts that the audio chunk length should match the return
+    /// of the get_samples_per_frame method.
+    /// 
+    /// Assumes sample rate match the configured for the detector.
+    /// 
+    /// Asserts that detector bits_per_sample is one of: 8, 16.
+    /// 
+    /// Asserts that detector sample_format is 'int'.
     pub fn process_i16(&mut self, audio_chunk: &[i16]) -> Option<DetectedWakeword> {
         assert!(self.bits_per_sample == 8 || self.bits_per_sample == 16);
         self.process_int(
@@ -343,6 +478,16 @@ impl WakewordDetector {
                 .collect::<Vec<_>>(),
         )
     }
+    /// Process i32 audio chunks.
+    /// 
+    /// Asserts that the audio chunk length should match the return
+    /// of the get_samples_per_frame method.
+    /// 
+    /// Assumes sample rate match the configured for the detector.
+    /// 
+    /// Asserts that detector bits_per_sample is one of: 8, 16, 24, 32.
+    /// 
+    /// Asserts that detector sample_format is 'int'.
     pub fn process_i32(&mut self, audio_chunk: &[i32]) -> Option<DetectedWakeword> {
         assert!(
             self.bits_per_sample == 8
@@ -352,6 +497,72 @@ impl WakewordDetector {
         );
         self.process_int(audio_chunk)
     }
+    /// Process f32 audio chunks.
+    /// 
+    /// Asserts that the audio chunk length should match the return
+    /// of the get_samples_per_frame method.
+    /// 
+    /// Assumes sample rate match the configured for the detector.
+    /// 
+    /// Asserts that detector bits_per_sample is 32.
+    /// 
+    /// Asserts that detector sample_format is 'float'.
+    pub fn process_f32(&mut self, audio_chunk: &[f32]) -> Option<DetectedWakeword> {
+        assert!(audio_chunk.len() == self.samples_per_frame);
+        assert!(self.bits_per_sample == 32);
+        assert!(self.sample_format == SampleFormat::Float);
+        let float_buffer: Vec<f32> = audio_chunk
+            .into_iter()
+            .map(|s| s * 32767.0)
+            .collect::<Vec<_>>();
+        let resampled_audio = if self.resampler.is_some() {
+            let resampler = self.resampler.as_mut().unwrap();
+            let waves_in = vec![float_buffer; 1];
+            let waves_out = self.resampler_out_buffer.as_mut().unwrap();
+            resampler
+                .process_into_buffer(&waves_in, waves_out, None)
+                .unwrap();
+            let result = waves_out.get(0).unwrap();
+            result.to_vec()
+        } else {
+            float_buffer
+        };
+        self.apply_vad_detection(resampled_audio)
+    }
+    /// Returns the desired chunk size.
+    pub fn get_samples_per_frame(&self) -> usize {
+        self.samples_per_frame
+    }
+    fn add_keyword_from_model(
+        &mut self,
+        model: WakewordModel,
+        enabled: bool,
+    ) -> Result<(), String> {
+        let keyword = model.keyword.clone();
+        let wakeword = Wakeword::from_model(model, enabled);
+        self.update_detection_frame_size(wakeword.get_min_frames(), wakeword.get_max_frames());
+        self.wakewords.insert(keyword, wakeword);
+        Ok(())
+    }
+    fn update_detection_frame_size(&mut self, min_frames: usize, max_frames: usize) {
+        self.min_frames = std::cmp::min(self.min_frames, min_frames);
+        self.max_frames = std::cmp::max(self.max_frames, max_frames);
+    }
+    fn get_wakeword_model(&self, name: &String) -> Result<WakewordModel, String> {
+        let wakeword = self.wakewords.get(name);
+        if wakeword.is_none() {
+            Err(String::from("Missing wakeword"))
+        } else {
+            let features = wakeword.unwrap().get_templates();
+            let model = WakewordModel::new(
+                name.clone(),
+                features,
+                wakeword.unwrap().get_averaged_threshold(),
+                wakeword.unwrap().get_threshold(),
+            );
+            Ok(model)
+        }
+    }
     fn process_int(&mut self, audio_chunk: &[i32]) -> Option<DetectedWakeword> {
         assert!(audio_chunk.len() == self.samples_per_frame);
         assert!(self.sample_format == SampleFormat::Int);
@@ -383,46 +594,33 @@ impl WakewordDetector {
         };
         self.apply_vad_detection(resampled_audio)
     }
-    pub fn process_f32(&mut self, audio_chunk: &[f32]) -> Option<DetectedWakeword> {
-        assert!(audio_chunk.len() == self.samples_per_frame);
-        assert!(self.bits_per_sample == 32);
-        assert!(self.sample_format == SampleFormat::Float);
-        let float_buffer: Vec<f32> = audio_chunk
-            .into_iter()
-            .map(|s| s * 32767.0)
-            .collect::<Vec<_>>();
-        let resampled_audio = if self.resampler.is_some() {
-            let resampler = self.resampler.as_mut().unwrap();
-            let waves_in = vec![float_buffer; 1];
-            let waves_out = self.resampler_out_buffer.as_mut().unwrap();
-            resampler
-                .process_into_buffer(&waves_in, waves_out, None)
-                .unwrap();
-            let result = waves_out.get(0).unwrap();
-            result.to_vec()
-        } else {
-            float_buffer
-        };
-        self.apply_vad_detection(resampled_audio)
-    }
     fn apply_vad_detection(&mut self, resampled_audio: Vec<f32>) -> Option<DetectedWakeword> {
-        if self.vad_detector.is_some()
-            && self.voice_detection_time.elapsed().unwrap().as_secs() >= self.vad_delay as u64
+        if self.result_state.is_none()
+            && self.vad_detector.is_some()
+            && (self.vad_enabled
+                || self.voice_detection_time.elapsed().unwrap().as_secs() >= self.vad_delay as u64)
         {
             let vad = self.vad_detector.as_mut().unwrap();
+            if !self.vad_enabled {
+                debug!("switching to vad detector");
+                self.vad_enabled = true;
+            }
             let is_voice_result = vad.is_voice_segment(
                 &resampled_audio
                     .iter()
                     .map(|i| *i as i16)
                     .collect::<Vec<i16>>(),
             );
-            if self.voice_detections.len() == 100 {
-                self.voice_detections.drain(0..1);
-            }
             self.voice_detections
                 .push(is_voice_result.is_err() || is_voice_result.unwrap());
+            if self.voice_detections.len() < 30 {
+                return self.process_encoded_audio(&resampled_audio, true);
+            }
+            if self.voice_detections.len() > 100 {
+                self.voice_detections.drain(0..1);
+            }
             if self.voice_detections.iter().filter(|i| **i == true).count()
-                >= (self.vad_sensitivity * 100.) as usize
+                >= (self.vad_sensitivity * self.voice_detections.len() as f32) as usize
             {
                 debug!("voice detected; processing cache");
                 self.audio_cache
@@ -434,12 +632,13 @@ impl WakewordDetector {
                     });
                 self.voice_detections.clear();
                 self.voice_detection_time = SystemTime::now();
-                debug!("detection time updated, processing last frame");
+                debug!("switching to feature detector");
+                self.vad_enabled = false;
                 self.process_encoded_audio(&resampled_audio, true)
             } else {
-                if self.audio_cache.len() >= self.min_frames {
+                if self.audio_cache.len() >= self.max_frames {
                     self.audio_cache
-                        .drain(0..=self.audio_cache.len() - self.min_frames);
+                        .drain(0..=self.audio_cache.len() - self.max_frames);
                 }
                 self.audio_cache.push(resampled_audio);
                 None
@@ -547,10 +746,15 @@ impl WakewordDetector {
         let result_option = self.get_best_keyword(features);
         match result_option {
             Some(result) => {
+                if self.vad_enabled {
+                    self.vad_enabled = false;
+                    debug!("switching to feature detector");
+                }
+                self.voice_detection_time = SystemTime::now();
                 if self.eager_mode {
                     if result.index != 0 {
                         debug!("Sorting '{}' templates", result.wakeword);
-                        self.keywords
+                        self.wakewords
                             .get_mut(&result.wakeword)
                             .unwrap()
                             .prioritize_template(result.index);
@@ -563,9 +767,10 @@ impl WakewordDetector {
                     Some(result)
                 } else {
                     if self.result_state.is_some() {
-                        let prev_wakeword = self.result_state.as_ref().unwrap().wakeword.clone();
-                        let prev_score = self.result_state.as_ref().unwrap().score;
-                        let prev_index = self.result_state.as_ref().unwrap().index;
+                        let prev_result = self.result_state.as_ref().unwrap();
+                        let prev_wakeword = prev_result.wakeword.clone();
+                        let prev_score = prev_result.score;
+                        let prev_index = prev_result.index;
                         if result.wakeword == prev_wakeword && result.score < prev_score {
                             debug!(
                                 "keyword '{}' detected, score {}",
@@ -583,7 +788,22 @@ impl WakewordDetector {
                     None
                 }
             }
-            None => None,
+            None => {
+                if self.result_state.is_some() {
+                    let prev_result = self.result_state.as_ref().unwrap();
+                    let wakeword = prev_result.wakeword.clone();
+                    let score = prev_result.score;
+                    let index = prev_result.index;
+                    self.reset();
+                    Some(DetectedWakeword {
+                        wakeword,
+                        score,
+                        index,
+                    })
+                } else {
+                    None
+                }
+            }
         }
     }
     fn reset(&mut self) {
@@ -621,49 +841,74 @@ impl WakewordDetector {
     }
     fn get_best_keyword(&self, features: Vec<Vec<f32>>) -> Option<DetectedWakeword> {
         let mut detections: Vec<DetectedWakeword> =
-            if self.single_thread || self.keywords.len() <= 1 {
-                self.keywords
+            if self.single_thread || self.wakewords.len() <= 1 {
+                self.wakewords
                     .iter()
-                    .filter_map(|(name, keyword)| {
-                        let templates = keyword.get_templates();
-                        let threshold = keyword.get_threshold().unwrap_or(self.threshold);
+                    .filter_map(|(name, wakeword)| {
+                        if !wakeword.is_enabled()
+                            || self.result_state.is_some()
+                                && self.result_state.as_ref().unwrap().wakeword != *name
+                        {
+                            return None;
+                        }
+                        let templates = wakeword.get_templates();
+                        let averaged_template = wakeword.get_averaged_template();
+                        let averaged_threshold = if self.result_state.is_none() {
+                            wakeword
+                                .get_averaged_threshold()
+                                .unwrap_or(self.averaged_threshold)
+                        } else {
+                            0.
+                        };
+                        let threshold = wakeword.get_threshold().unwrap_or(self.threshold);
                         run_wakeword_detection(
-                            self.comparator_band_size,
-                            self.comparator_ref,
+                            &self.comparator,
                             &templates,
+                            averaged_template,
                             self.eager_mode,
                             &features,
+                            averaged_threshold,
                             threshold,
                             name.clone(),
                         )
                     })
                     .collect::<Vec<_>>()
             } else {
-                self.keywords
+                self.wakewords
                     .iter()
-                    .filter_map(|(name, keyword)| {
-                        if !keyword.is_enabled() {
-                            None
-                        } else {
-                            let wakeword_name = name.clone();
-                            let threshold = keyword.get_threshold().unwrap_or(self.threshold);
-                            let templates = keyword.get_templates();
-                            let comparator_band_size = self.comparator_band_size;
-                            let comparator_ref = self.comparator_ref;
-                            let eager_mode = self.eager_mode;
-                            let features_copy = features.to_vec();
-                            Some(thread::spawn(move || {
-                                run_wakeword_detection(
-                                    comparator_band_size,
-                                    comparator_ref,
-                                    &templates,
-                                    eager_mode,
-                                    &features_copy,
-                                    threshold,
-                                    wakeword_name,
-                                )
-                            }))
+                    .filter_map(|(name, wakeword)| {
+                        if !wakeword.is_enabled()
+                            || self.result_state.is_some()
+                                && self.result_state.as_ref().unwrap().wakeword != *name
+                        {
+                            return None;
                         }
+                        let wakeword_name = name.clone();
+                        let threshold = wakeword.get_threshold().unwrap_or(self.threshold);
+                        let averaged_threshold = if self.result_state.is_none() {
+                            wakeword
+                                .get_averaged_threshold()
+                                .unwrap_or(self.averaged_threshold)
+                        } else {
+                            0.
+                        };
+                        let templates = wakeword.get_templates();
+                        let averaged_template = wakeword.get_averaged_template();
+                        let comparator = self.comparator.clone();
+                        let eager_mode = self.eager_mode;
+                        let features_copy = features.to_vec();
+                        Some(thread::spawn(move || {
+                            run_wakeword_detection(
+                                &comparator,
+                                &templates,
+                                averaged_template,
+                                eager_mode,
+                                &features_copy,
+                                averaged_threshold,
+                                threshold,
+                                wakeword_name,
+                            )
+                        }))
                     })
                     .map(JoinHandle::join)
                     .filter_map(Result::unwrap)
@@ -676,9 +921,6 @@ impl WakewordDetector {
             Some(detections.remove(0))
         }
     }
-    pub fn get_samples_per_frame(&self) -> usize {
-        self.samples_per_frame
-    }
 }
 fn resample_audio(input_sample_rate: usize, audio_pcm_signed: &[f32]) -> Vec<f32> {
     let mut resampler =
@@ -698,20 +940,17 @@ fn resample_audio(input_sample_rate: usize, audio_pcm_signed: &[f32]) -> Vec<f32
 }
 
 fn run_wakeword_detection(
-    comparator_band_size: usize,
-    comparator_ref: f32,
+    comparator: &FeatureComparator,
     templates: &[Vec<Vec<f32>>],
+    averaged_template: Option<Vec<Vec<f32>>>,
     eager_mode: bool,
     features: &[Vec<f32>],
+    averaged_threshold: f32,
     threshold: f32,
     wakeword_name: String,
 ) -> Option<DetectedWakeword> {
-    let comparator = FeatureComparator::new(comparator_band_size, comparator_ref);
-    let mut detection: Option<DetectedWakeword> = None;
-    for (index, template) in templates.iter().enumerate() {
-        if eager_mode && detection.is_some() {
-            break;
-        }
+    if averaged_threshold > 0. && averaged_template.is_some() {
+        let template = averaged_template.unwrap();
         let mut frames = features.to_vec();
         if frames.len() > template.len() {
             frames.drain(template.len()..frames.len());
@@ -720,6 +959,25 @@ fn run_wakeword_detection(
             &template.iter().map(|item| &item[..]).collect::<Vec<_>>(),
             &frames.iter().map(|item| &item[..]).collect::<Vec<_>>(),
         );
+        if score < averaged_threshold {
+            return None;
+        }
+        debug!(
+            "wakeword '{}' passes averaged detection with score {}",
+            wakeword_name, score
+        );
+    }
+    let mut detection: Option<DetectedWakeword> = None;
+    for (index, template) in templates.iter().enumerate() {
+        let mut frames = features.to_vec();
+        if frames.len() > template.len() {
+            frames.drain(template.len()..frames.len());
+        }
+        let score = comparator.compare(
+            &template.iter().map(|item| &item[..]).collect::<Vec<_>>(),
+            &frames.iter().map(|item| &item[..]).collect::<Vec<_>>(),
+        );
+        debug!("wakeword '{}' scored {}", wakeword_name, score);
         if score < threshold {
             continue;
         }
@@ -732,16 +990,19 @@ fn run_wakeword_detection(
             score,
             index,
         });
+        if eager_mode {
+            break;
+        }
     }
     detection
 }
-
+/// Represents a successful wakeword detection.
 pub struct DetectedWakeword {
-    // Detected wakeword name
+    /// Detected wakeword name.
     pub wakeword: String,
-    // Detection score
+    /// Detection score.
     pub score: f32,
-    // Detected wakeword template index
+    /// Detected wakeword template index.
     pub index: usize,
 }
 impl Clone for DetectedWakeword {
